import os
import re
from collections import defaultdict

import numpy as np

from .rules import *


def generate_guards(templates, temp_fname, guard_fname="gen_guards.txt"):
    guards = defaultdict(list)
    temp_fname = os.path.splitext(os.path.basename(temp_fname))[0]
    negate = False
    for i, temp in enumerate(templates):
        guard_set = []
        pos_tag, morph, has_aux = [], [], []

        for p in temp['props']:
            pos_tag.append(p['pos'])
            has_aux.append(p['has_aux'])
            morph.append(str(p['feats']).strip())

        pos_tag, morph = set(pos_tag), set(morph)

        has_aux = all(has_aux)
        for pos, mrph in zip(pos_tag, morph):
            guard = []
            if pos:
                guard.append(f"n.pos is {pos}")
            if mrph:
                guard.append(f"n.morph has {mrph}")
            guard_set.append(guard)

        guard = []
        arity, aux_not_exists = defaultdict(set), None
        for j, clause in enumerate([temp['question'], temp['answer']]):
            for el in clause.strip().split():
                if el.startswith('[w.lemma'):
                    continue

                if el.endswith('.lemma]'):
                    el = el.replace('.lemma', '')


                # TODO: remove the dependency on the template word being called `w` (make it an argument)
                if el == '[w]' and pos_tag == 'VERB' and not has_aux and aux_not_exists != 'NA':
                    aux_not_exists = 'INS'

                if el.startswith('[w.aux') or el.startswith('<w.aux'):
                    aux_not_exists = 'NA'

                # ignore negations, cuz they're not requirements
                if el.endswith(">") and negate:
                    negate = False
                    continue
                if negate: continue
                if el.strip() == "-":
                    negate = True
                    continue

                parts = el.split("-")[0].strip().split(".")
                if len(parts) >= 2:
                    gg = "n"
                    for part in parts[1:]:
                        p = re.sub(r'#[0-9]+', '', part.strip())
                        gg += ".{}".format(p.replace('>', '').replace(']', ''))
                    arity[gg].add(".".join(parts[1:]).replace('>', '').replace(']', ''))
                    if (parts[1].strip().startswith('nsubj') and len(parts) == 2) or (j == 1 and parts[-1].strip().startswith('nsubj')):
                        # either the clause has nsubj as a direct descendant of root
                        # or the answer points to subject of any clause in the sentence
                        guard.insert(0, f'{gg}.morph has_not PronType=Rel')

        for g, a in arity.items():
            N = len(a)
            if N > 1:
                guard.append(f"{g}/{N} exists")
            else:
                guard.append(f"{g} exists")

        if aux_not_exists == 'INS':
            guard.append('n.aux not_exists')

        for g in guard_set:
            new_guard = ", ".join(g + guard)
            guards[new_guard.strip()].append("{}{}".format(temp_fname, i+1))

    with open(guard_fname, "w") as f:
        for g, t in guards.items():
            f.write("{} -> {}\n".format(g, ",".join(t)))
    return guards



class GuardTreeNode:
    def __init__(self, rule, cond):
        self.__cond = cond
        self.__rule = rule
        self.__children = dict()
        self.templates = []
        self.parent = None

    def add_child(self, c):
        if c.rule in self.children:
            obj = self.children[c.rule]
            # this is just a safety code, since this should not happen
            for ch in obj.children.values():
                c.add_child(ch)
            c.templates.extend(obj.templates)
        c.parent = self
        self.children[c.rule] = c

    @property
    def children(self):
        return self.__children

    def is_root(self):
        return self.__rule == "ROOT" and self.__cond is None

    @property
    def rule(self):
        return self.__rule

    def cond(self, n, check_subtrees=False):
        return self.__cond(n, check_subtrees)

    @classmethod
    def init(cls, gtn):
        return cls(gtn.rule, gtn.cond)

    def print_tree(self, indent=0):
        print(" " * indent + str(self))
        for c in self.__children.values():
            c.print_tree(indent + 2)


    def __str__(self):
        return f"{self.__rule}___{(','.join(self.templates) or 'no_templates')}___{len(self.__children)}"


def load_guards(files):
    def pair(g_from, g_to, templates=None):
        g_to = GuardTreeNode.init(g_to)
        if templates:
            g_to.templates.extend(templates)
        g_from.add_child(g_to)
        g_from = g_to
        return g_from

    # we assume that we get udon2.Node as an input for every guard
    guard_nodes, guard_freq, parsed_guards = {}, defaultdict(int), []
    for fname in files:
        with open(fname) as f:
            for rule in f:
                guard = parse(rule, Guard)
                parsed_guards.append(guard)

                for ub in guard.uclauses:
                    guard_rule, guard_cond = str(ub), None
                    if guard_rule not in guard_nodes:
                        if ub.cond == 'exists':
                            if hasattr(ub, 'arity'):
                                try:
                                    arity = int(ub.arity)
                                except:
                                    continue
                                # Need to fix this for checking subtrees
                                guard_cond = lambda n,cs,v=ub.tag,a=arity: check_arity(n, v, a, check_subtrees=cs)
                            else:
                                guard_cond = lambda n,cs,v=ub.tag: relation_exists(n, v, check_subtrees=cs)
                        elif ub.cond == 'not_exists':
                            guard_cond = lambda n,cs,v=ub.tag: relation_not_exists(n, v, check_subtrees=cs)

                        if guard_cond:
                            gtn = GuardTreeNode(guard_rule, guard_cond)
                            guard_nodes[guard_rule] = gtn
                    guard_freq[guard_rule] += 1

                for gb in guard.bclauses:
                    guard_rule, guard_cond = str(gb), None

                    tag_chain_cond = lambda x,cs: x
                    if gb.tag:
                        tag_chain_cond = lambda x,cs,v=gb.tag: filter_by_rel_chain(x, v, check_subtrees=cs)

                    if guard_rule not in guard_nodes:
                        if gb.prop == 'pos':
                            if gb.cond == 'is':
                                guard_cond = lambda n,cs,v=gb.val,f=tag_chain_cond: filter_by_pos(n, v, f=f, check_subtrees=cs)
                            elif gb.cond == 'is_not':
                                guard_cond = lambda n,cs,v=gb.val,f=tag_chain_cond: filter_by_pos(n, v, f=f, check_subtrees=cs, negate=True)
                            elif gb.cond == 'is_in':
                                pass
                        # elif gb.prop == 'rel':
                        #     if gb.cond == 'is':
                        #         guard_cond = lambda n,v=gb.val,f=tag_chain_cond: filter_by_rel(f(n), v)
                        #     elif gb.cond == 'is_not':
                        #         guard_cond = lambda n,v=gb.val,f=tag_chain_cond: filter_by_rel(f(n), v, negate=True)
                        #     elif gb.cond == 'is_in':
                        #         pass
                        elif gb.prop == 'morph':
                            if gb.cond == 'has':
                                guard_cond = lambda n,cs,v=gb.val,f=tag_chain_cond: filter_by_morph(n, v, f=f, check_subtrees=cs)
                            elif gb.cond == 'has_not':
                                guard_cond = lambda n,cs,v=gb.val,f=tag_chain_cond: filter_by_morph(n, v, f=f, check_subtrees=cs, negate=True)
                        if guard_cond:
                            gtn = GuardTreeNode(guard_rule, guard_cond)
                            guard_nodes[guard_rule] = gtn
                    guard_freq[guard_rule] += 1

    # print(guard_freq)
    direct_children = set()
    for j, p in enumerate(parsed_guards):
        clauses = list(set(map(str, p.uclauses))) + list(set(map(str, p.bclauses)))

        if len(clauses) == 1:
            g = GuardTreeNode.init(guard_nodes[clauses[0]])
            g.templates.extend(p.templates.split(','))
            direct_children.add(g)
        else:
            freqs = [guard_freq[x] for x in clauses]
            ind = np.argsort(freqs)[::-1] # reverse argsort

            direct_children.add(guard_nodes[clauses[ind[0]]])
            if len(clauses) > 2:
                g_from = pair(guard_nodes[clauses[ind[0]]], guard_nodes[clauses[ind[1]]])
                for i in ind[2:-1]:
                    g_from = pair(g_from, guard_nodes[clauses[i]])
                g_from = pair(g_from, guard_nodes[clauses[ind[-1]]], p.templates.split(','))
            else:
                g_from = pair(guard_nodes[clauses[ind[0]]], guard_nodes[clauses[ind[1]]], p.templates.split(','))

    root = GuardTreeNode("ROOT", None)
    for gn in direct_children:
        root.add_child(gn)
    return root


# -*- coding: utf-8 -*-
import os
import re
import json
from collections import defaultdict
from itertools import product, combinations
from pypeg2 import *

from .common import chain_to_root


telugu_chars = '\u0C00-\u0C7F'
arabic_chars = '\u0600-\u06FF\ufb50-\ufdff\ufe70-\ufefc'
japanese_chars = '\u3000-\u303f\u3040-\u309f\u30a0-\u30ff\uff00-\uff9f\u4e00-\u9faf\u3400-\u4dbf'
question_word = re.compile(r'[{0}{1}{2}\w\?,\.]+(\'/[{0}{1}{2}\w\?,\.]+)*'.format(japanese_chars, arabic_chars, telugu_chars), re.U)
inter_word = re.compile(r'[{0}{1}{2}\w0-9\'\-\?,\.]+'.format(japanese_chars, arabic_chars, telugu_chars), re.U)


# to be fixed later with UDon2
base_folder = os.path.dirname(__file__)
lang_spec = json.load(open(os.path.join(base_folder, 'lang_spec_feats.json')))
lang_spec_tags = [x for d in lang_spec.values() for x in d.keys()]
lang_spec_vals = [y for d in lang_spec.values() for x in d.values() for y in x]


class PosTag(Keyword):
    """
    A keyword (enum), describing a part-of-speech tag, should be capitalized
    """
    VALUES = ["ADJ", "ADP", "PUNCT", "ADV", "AUX", "SYM", "INTJ", "CCONJ", "X", "NOUN", "DET", "PROPN", "NUM", "VERB", "PART", "PRON", "SCONJ"]
    grammar = Enum(*(K(v) for v in VALUES))


class MorphTag(Keyword):
    VALUES = ["PronType", "Gender", "VerbForm", "NumType", "NumForm", "Animacy", "Mood", "Poss", "NounClass", "Tense", "Reflex", "Number", "Aspect",
              "Case", "Voice", "Abbr", "Definite", "Evident", "Typo", "Degree", "Polarity", "Person", "Polite", "Clusivity", "Variant",
              "Derivation", "Foreign", "InfForm", "PartForm", "Person[psor]", "Number[psor]"] + lang_spec_tags
    grammar = Enum(*(K(v) for v in VALUES))
    regex = re.compile(r'[\w\[\]]+', re.U)


class MorphValue(Keyword):
    VALUES = ['Ex', 'Bantu16', 'Ben', 'Pass', 'Dis', 'Hab', 'Cau', 'Ger', 'Qot', 'Abs', 'Neg', 'Cons', 'Bantu18', 'Humb', 'Gen', 'Cmp', 'Mult',
              'Sing', 'Cns', 'Tra', 'Part', 'Nom', 'Hum', 'All', 'Pauc', 'Fem', 'Masc', 'Bantu8', 'Sup', 'Ess', 'Coll', 'Acc', 'Conv', 'Abe',
              'Prp', 'Dat', 'Grpa', 'Count', 'Equ', 'Infm', 'Jus', 'Inf', 'Adm', '1', 'Abl', 'Gdv', 'Com', 'Sets', 'Rel', 'Prosp', 'Dir',
              'Bantu13', '4', 'Fh', 'Pres', 'Bantu20', 'Int', 'Rcp', 'Per', 'Bantu2', 'Ade', 'Tot', 'Ptan', 'Form', 'Plur', 'Ter', 'Loc', 'Bantu1',
              'Frac', 'Anim', 'Bantu17', '0', 'Dual', 'Pot', 'Iter', 'Bantu15', 'Bantu11', 'Opt', 'Range', 'Fin', '3', 'Bantu3', 'Elev', 'Exc',
              '2', 'Dist', 'Bantu5', 'Prog', 'Pqp', 'Vnoun', 'Bantu4', 'Nfh', 'Emp', 'In', 'Pos', 'Des', 'Inan', 'Ord', 'Del', 'Add', 'Def', 'Ine',
              'Bantu10', 'Act', 'Bantu19', 'Ill', 'Imp', 'Art', 'Inv', 'Voc', 'Neut', 'Ind', 'Bantu14', 'Mid', 'Dem', 'Par', 'Past', 'Card', 'Fut',
              'Prs', 'Grpl', 'Lat', 'Antip', 'Perf', 'Yes', 'Bantu7', 'Tem', 'Ela', 'Spec', 'Nec', 'Tri', 'Ins', 'Cnd', 'Erg', 'Bantu9', 'Bantu12',
              'Nhum', 'Bantu6', 'Sub', "Short", "Inen", "Ja", "Lainen", "Llinen", "Vs", "U", "Tar", "Digit"] + lang_spec_vals
    grammar = Enum(*(K(v) for v in VALUES))


class MorphValueChain(List):
    grammar = MorphValue, maybe_some(",", MorphValue)

    def __str__(self):
        return "|".join(map(str, list(self)))


class MorphKeyVal:
    grammar = attr('tag', MorphTag), '=', attr('val', MorphValueChain)

    def __str__(self):
        return f"{self.tag}={self.val}"


class MorphKeyValChain(List):
    grammar = MorphKeyVal, maybe_some("|", MorphKeyVal)

    def __str__(self):
        return "|".join(map(str, list(self)))


class Prop(Keyword):
    VALUES = ['pos', 'rel', 'morph', 'text', 'lemma']
    grammar = Enum(*(K(v) for v in VALUES))


class DepRelTag(Keyword):
    CORE_NOMINALS = ["nsubj", "obj", "iobj"]
    CORE_CLAUSES = ["csubj", "ccomp", "xcomp"]
    NON_CORE_NOMINALS = ["obl", "vocative", "expl", "dislocated"]
    NON_CORE_CLAUSES = ["advcl"]
    NON_CORE_MODIFIERS = ["advmod", "discourse"]
    NON_CORE_FUNCTION = ["aux", "cop", "mark"]
    NOMINAL_DEP_NOMINALS = ["nmod", "appos", "nummod"]
    NOMINAL_DEP_CLAUSES = ["acl"]
    NOMINAL_DEP_MODIFIERS = ["amod"]
    NOMINAL_DEP_FUNCTION = ["det", "clf", "case"]

    COORDINATION = ["conj", "cc"]
    MWE = ["fixed", "flat", "compound"]
    LOOSE = ["list", "parataxis"]
    SPECIAL = ["orphan", "goeswith", "reparandum"]
    OTHER = ["punct", "root", "dep"]

    VALUES = CORE_NOMINALS + CORE_CLAUSES + NON_CORE_NOMINALS + NON_CORE_CLAUSES + \
            NON_CORE_MODIFIERS + NON_CORE_FUNCTION + NOMINAL_DEP_FUNCTION + NOMINAL_DEP_MODIFIERS + \
            NOMINAL_DEP_CLAUSES + NOMINAL_DEP_NOMINALS + COORDINATION + MWE + LOOSE + SPECIAL + OTHER
    grammar = Enum(*(K(v) for v in VALUES))


class DepRelModifier(Keyword):
    grammar = Enum(K("pass"), K("relcl"), K("agent"), K("prt"), K("name"), K("poss"), K("npmod"), K("ds"), K("gobj"), K("emph"),
                   K("tmod"), K("preconj"), K("foreign"), K("predet"), K("cleft"), K("gov"), K("cop"), K("nn"), K("own"), K("arg"),
                   K("svc"), K("nc"), K("gsubj"))


class ModifiedDepRelTag:
    grammar = attr('rel', DepRelTag), optional(':', attr('mod', DepRelModifier)), optional('#', attr('id', str)), optional(flag("exact", '*'))

    def __str__(self):
        s = self.rel
        if hasattr(self, 'mod'):
            s += ":" + self.mod
        if hasattr(self, 'exact') and self.exact:
            s += '*'
        return s


class DepTagChain(List):
    grammar = maybe_some(".", ModifiedDepRelTag)

    def __str__(self):
        return ".".join([str(a) for a in self])


class NegativeDepEntity:
    grammar = attr('tag', ModifiedDepRelTag), attr('chain', optional(DepTagChain))

    def __str__(self):
        return str(self.tag) + (("." + str(self.chain)) if hasattr(self, 'chain') and len(self.chain) > 0 else "")


class NegativeDepRel(List):
    grammar = maybe_some(optional(blank), '-', optional(blank), NegativeDepEntity)

    def __str__(self):
        return " - ".join([str(r) for r in self])


class UnaryCondition(Keyword):
    grammar = Enum(K("exists_once"), K("exists"), K("not_exists"))


class UnaryClause(Namespace):
    grammar = name(), attr('tag', DepTagChain), optional('/', attr('arity', str)), optional(blank), attr("cond", UnaryCondition)

    def __str__(self):
        return self.name + "." + ".".join([str(t) for t in list(self.tag)]) +\
            (f"/{self.arity}" if hasattr(self, 'arity') else "") +  " " + str(self.cond)


class BinaryCondition(Keyword):
    grammar = Enum(K("is"), K("is_not"), K("is_in"), K("has"), K("has_not"))


class BinaryClause(Namespace):
    grammar = name(), attr('tag', DepTagChain), optional('.', attr('prop', Prop)), optional(blank), attr('cond', BinaryCondition),\
        optional(blank), attr('val', [PosTag, DepRelTag, MorphKeyValChain])

    def __str__(self):
        return self.name + ("." + ".".join([str(t) for t in list(self.tag)]) if self.tag else "") + \
                "." + self.prop + " " + self.cond + " " + str(self.val)

class Guard(Namespace):
    grammar = attr('bclauses', maybe_some(csl(BinaryClause))), optional(optional(","), attr('uclauses', maybe_some(csl(UnaryClause)))),\
        optional(blank), "->", optional(blank), attr('templates', restline)

class SubtreePlaceholder:
    grammar = '<', name(), optional(attr('tag', DepTagChain)), optional(attr('neg', NegativeDepRel)), '>'

    def __str__(self):
        return "<" + self.name + "." + ".".join([str(t) for t in list(self.tag)]) + (" - " + str(self.neg) if hasattr(self, 'neg') and len(self.neg) > 0 else "") + ">"


class NodePlaceholder:
    grammar = '[', name(), optional(attr('tag', DepTagChain)), optional(".", attr('prop', Prop)), ']'

    def __str__(self):
        return "[" + self.name + "." + ".".join([str(t) for t in list(self.tag)]) + "]"

class Template(List):
    grammar = some([question_word, inter_word, NodePlaceholder, SubtreePlaceholder]), optional(blank), "=>",\
        optional(blank), attr('answer', some([inter_word, NodePlaceholder, SubtreePlaceholder]))


def filter_by_pos(node, value, f=None, check_subtrees=False, negate=False):
    def __filter_by_pos(n):
        result = []
        pos_match = n.upos == value
        if (pos_match and (not negate)) or ((not pos_match) and negate):
            result.append(n)


        if f is not None and identify_fn:
            # if we don't have any tags before we might possibly want to check the subtrees
            # otherwise they have been checked by `filter_by_rel_chain`, so all that's left 
            # is to check POS-tags on the nodes themselves, which we have already done above.
            if check_subtrees:
                # TOFIX: selectByPos returns udon2.NodeList, but want just a python list
                result.extend(n.select_by("upos", value, negate)) # search in the subtree except the POS
        return result

    identify_fn = True
    if f is not None:
        identity_fn = f(node, check_subtrees) == node
        node = f(node, check_subtrees)
    if type(node) == list:
        # see comment in filter_by_morph
        res = []
        for beg, end in node:
            filtered = __filter_by_pos(end)
            if filtered:
                res.append(beg)
    else:
        res = __filter_by_pos(node)
    return res
        

# TODO: when filtering by rel, recall that if node's rel satsifies the condition, then we want to have its parent as `n`
# def filter_by_rel(node, value, negate=False):
#     def __filter_by_rel(n, check_subtrees=True):
#         result = []
#         rel_match = n.get_rel() == value
#         if (rel_match and (not negate)) or ((not rel_match) and negate):
#             result.append(n)
        
#         if check_subtrees:
#             # TOFIX: selectByPos returns udon2.NodeList, but want just a python list
#             if negate:
#                 result.extend(n.select_except_rel(value)) # search in the subtree
#             else:
#                 result.extend(n.select_by_rel(value)) # search in the subtree 
#         return result

#     if type(node) == list:
#         # see comment in filter_by_morph
#         res = []
#         for nd in node:
#             res.extend(__filter_by_rel(nd, check_subtrees=False))
#     else:
#         res = __filter_by_rel(node)
#     return res

def filter_by_rel_chain(node, rel_chain, check_subtrees=True):
    # TODO: make negate work
    cand, res = [(node, node)], []
    for i, r in enumerate(rel_chain):
        for beg, end in cand:
            if i == 0 and check_subtrees:
                for nd in end.select_by("deprel", str(r)): # search in the subtree
                    res.append((nd.parent, nd))
            else:
                for nd in end.get_by("deprel", str(r)): # not search in the subtree
                    res.append((beg, nd))
        cand, res = res, []
    return cand


def check_arity(node, value, arity, check_subtrees=False):
    freq, cand, node2id = defaultdict(int), {}, {}
    idx = 0
    for beg, end in filter_by_rel_chain(node, value, check_subtrees=check_subtrees):
        el_id = node2id.get(str(beg))
        if el_id is None:
            node2id[str(beg)] = idx
            el_id = idx
            idx += 1
        cand[el_id] = beg
        freq[el_id] += 1
    return [cand[idx] for idx, f in freq.items() if ((check_subtrees and cand[idx] == node) or not check_subtrees) and f >= arity]


def filter_by_morph(node, value, f=None, check_subtrees=False, negate=False):
    def __filter_by_morph(n):
        result = []
        # value is MorphKeyValChain
        v = str(value)

        node_has_all_morph = n.has_all("feats", v)
        if (node_has_all_morph and (not negate)) or ((not node_has_all_morph) and negate):
            result.append(n)

        if f is not None and identify_fn:
            # if we don't have any tags before we might possibly want to check the subtrees
            # otherwise they have been checked by `filter_by_rel_chain`, so all that's left 
            # is to check POS-tags on the nodes themselves, which we have already done above.
            if check_subtrees:
                result.extend(n.select_having("feats", v, negate)) # search in the subtree
        return result

    identify_fn = True
    if f is not None:
        identity_fn = f(node, check_subtrees) == node
        node = f(node, check_subtrees)
    
    if type(node) == list:
        # this means we've invoked a chain before, e.g.:
        # w.acl.nsubj.morph has_not PronType=Rel
        # This will select all w.acl.nsubj and then we efectively will want just to check those nodes
        # for the given condition and not the whole subtrees.
        res = []
        for beg, end in node:
            filtered = __filter_by_morph(end)
            if filtered:
                res.append(beg)

    else:
        filtered = __filter_by_morph(node)
        res = [node] if filtered else None
    return res


def relation_exists(node, value, check_subtrees=False):
    if len(value) == 1:
        if check_subtrees:
            return [x.parent for x in node.select_by("deprel", str(value[0]))]
        else:
            if node.child_has_prop("deprel", str(value[0])):
                return [node]
            else:
                return []
    else:
        res = filter_by_rel_chain(node, value, check_subtrees=check_subtrees)
        # TODO: fix this weird if statement
        return [beg for beg, end in res if (check_subtrees and beg == node) or not check_subtrees]


def relation_not_exists(node, value, check_subtrees=False):
    nodes = relation_exists(node, value, check_subtrees)

    if check_subtrees:
        return [x for x in node.get_subtree_nodes() if x not in nodes]
    else:

        return [] if nodes else [node]


def prop2val(node, prop):
    if prop == 'lemma':
        return node.lemma
    elif prop == 'text':
        return node.form
    elif prop == 'pos':
        return node.upos
    elif prop == 'morph':
        return str(node.feats)
    elif prop == 'rel':
        return node.deprel


class TemplateInstance:
    def __init__(self, cand, text=[]):
        self.__chain = chain_to_root(cand, only_ids=True) if cand else None
        self.__text = set()
        for t in text:
            self.__text.add(t)

    def add_text(self, text):
        self.__text.add(text)

    @property
    def text(self):
        return self.__text

    @property
    def chain(self):
        return self.__chain

    @property
    def is_empty(self):
        return bool(self.__text)


def placeholder2lambda(node, pholder):
    def node2text(n):
        if type(pholder) == NodePlaceholder:
            return prop2val(n, pholder.prop) if hasattr(pholder, 'prop') else n.form
        elif type(pholder) == SubtreePlaceholder:
            return n.get_subtree_text()
        else:
            return None


    if pholder.name == 'qw':
        # this is a question word, we leave it like that
        el = TemplateInstance(None)
        el.add_text(pholder.name)
        return {None: el}
    elif hasattr(pholder, 'tag') and pholder.tag:
        node.reset_subtree()
        candidates = node.get_by_deprel_chain(".".join([str(tag) for tag in pholder.tag]))
        last_tag = pholder.tag[-1]

        # print(".".join([str(tag) for tag in pholder.tag]))
        # print("PHOLDER", pholder, type(pholder), len(candidates), "<--", node.form, " (", node.get_subtree_text(), ")")

        if candidates:
            fixed_cand = {}
            for cand in candidates:
                cand.reset_subtree()
                if hasattr(pholder, 'neg'):
                    if len(pholder.neg) > 0:
                        options = []
                        # make sure the exact number of things is being ignored, i.e. if there are two advcls in pholder.neg and
                        # two of them can be found, just ignore them right away without adding them to options.

                        to_ignore = {}
                        for t in pholder.neg:
                            key = str(t)
                            if key not in to_ignore:
                                to_ignore[key] = [1, t]
                            else:
                                to_ignore[key][0] += 1

                        for str_chain, value in to_ignore.items():
                            repeats, t = value

                            if len(t.chain) > 0:
                                last_el = t.chain[-1]
                            else:
                                last_el = t.tag

                            if last_el.exact:
                                # this is to correctly parse templates like <w.advcl#8 - advcl#8* - advcbl#9>
                                # One should understand that if t is `advcl#8*`, then the same node
                                # as the current template root should simply be ignored
                                if last_tag.id == last_el.id:
                                    cand.ignore()
                                    continue

                                res_by_chain = cand.get_by_deprel_chain(str_chain[:-1])
                                N = len(res_by_chain)
                                
                                if N > 0:
                                    if N <= repeats:
                                        for r_by_chain in res_by_chain:
                                            r_by_chain.ignore()
                                    else:
                                        # we have more results than we need to remove
                                        options.append([(combo, False) for combo in combinations(res_by_chain, repeats)])
                                elif cand.deprel == str_chain[:-1]:
                                    cand.ignore()
                            else:
                                res_by_chain = cand.get_by_deprel_chain(str_chain)
                                N = len(res_by_chain)
                                if N > 0:
                                    if N <= repeats:
                                        for r_by_chain in res_by_chain:
                                            r_by_chain.ignore_subtree()
                                    else:
                                        # we have more results than we need to remove
                                        options.append([(combo, True) for combo in combinations(res_by_chain, repeats)])
                                elif cand.deprel == str_chain:
                                    cand.ignore_subtree()

                        if options:
                            cand_idx, cand_limit = 0, 50
                            fixed_cand[cand.id] = TemplateInstance(cand)
                            for option in product(*options):
                                cand_idx += 1
                                if cand_idx >= cand_limit: break # computational simplification
                                for nodes, subtree in option:
                                    for node in nodes:
                                        if subtree:
                                            node.ignore_subtree()
                                        else:
                                            node.ignore()
                                fixed_cand[cand.id].add_text(node2text(cand))
                                for nodes, subtree in option:
                                    for node in nodes:
                                        if subtree:
                                            node.reset_subtree()
                                        else:
                                            node.reset()
                if cand.id not in fixed_cand:
                    fixed_cand[cand.id] = TemplateInstance(cand)
                    fixed_cand[cand.id].add_text(node2text(cand))
            return fixed_cand
        else:
            # many of these occur for some reason!
            print("HERE2", pholder)
            return None
    elif pholder.name == 'w':
        el = TemplateInstance(node)
        if type(pholder) == NodePlaceholder:
            el.add_text(prop2val(node, pholder.prop) if hasattr(pholder, 'prop') else node.form)
            return {node.id: el}
        elif type(pholder) == SubtreePlaceholder:
            el.add_text(node.get_subtree_text())
            return {node.id: el}
        else:
            print("HERE3")
            return None
    else:
        el = TemplateInstance(None)
        el.add_text(str(pholder.name))
        return {None: el}


class SharedElement:
    def __init__(self, q_id, a_id, q_last, a_last):
        self.__q_id = q_id
        self.__a_id = a_id
        self.__last_in_chain = {
            'q': q_last,
            'a': a_last
        }

    @property
    def q(self):
        return self.__q_id

    @property
    def a(self):
        return self.__a_id

    def last_in_chain(self, k):
        return self.__last_in_chain.get(k, None)

    def __str__(self):
        return f"q{self.__q_id}:{self.__last_in_chain['q']}/a{self.__a_id}:{self.__last_in_chain['a']}"


def load_templates(files):
    templates = {}
    for fname in files:
        name_without_ext = os.path.splitext(os.path.basename(fname))[0]
        with open(fname, encoding='utf8') as f:
            for i, temp in enumerate(f):
                template = parse(temp.strip(), Template)
                tags = defaultdict(int)

                answer_ids = {
                    t.id: {'id': i, 'last': j == len(tmpl.tag) - 1}
                    for i, tmpl in enumerate(template.answer)
                    for j, t in enumerate(tmpl.tag if type(tmpl) != str else []) if hasattr(t, 'id')
                }

                templates[f'{name_without_ext}{i+1}'.strip()] = {
                    'wireframe': [type(x) for x in template],
                    'question': lambda n,t=template: [placeholder2lambda(n, x)
                        if type(x) in [SubtreePlaceholder, NodePlaceholder] else {None: TemplateInstance(None, [x])} for x in t],
                    'answer': lambda n,v=template.answer: [placeholder2lambda(n, x)
                        if type(x) in [SubtreePlaceholder, NodePlaceholder] else {None: TemplateInstance(None, [x])} for x in v],
                    'template': temp.strip(),
                    'shared_tags': [
                        SharedElement(i, answer_ids[t.id]['id'], j == len(tmpl.tag) - 1, answer_ids[t.id]['last']) 
                        for i, tmpl in enumerate(template)
                        for j, t in enumerate(tmpl.tag if type(tmpl) != str else []) if hasattr(t, 'id') and t.id in answer_ids
                    ]
                }
    return templates


def load_template_examples(files):
    examples = {}
    for fname in files:
        with open(fname) as f:
            for line in f:
                if line.startswith("id:"):
                    _, temp_id = line.split(": ")
                    examples[temp_id.strip()] = []
                elif line.strip():
                    s, q, a = line.strip().split(" | ")
                    examples[temp_id.strip()].append({
                        "sentence": s,
                        "question": q,
                        "answer": a
                    })
    return examples


if __name__ == '__main__':
    load_templates(['tmpl/en/templates/gen_templates.txt'])


import os
import json
import operator
import jsonlines
from collections import defaultdict

from pprint import pprint


class TextinatorLoader:
    @staticmethod
    def get_correct_answer(choices):
        return list(filter(lambda x: x['type'] == 'Correct answer', choices))[0]

    @staticmethod
    def from_files(fnames, lang='en'):
        answer_transformations = {}
        total, commented = 0, 0
        for fn in fnames:
            d = json.load(open(fn))
            assert 'data' in d, "Not compatible with Textinator format"

            for dp in d['data']:
                context, question, choices = dp['context'], dp['question'], dp['choices']
                ca = TextinatorLoader.get_correct_answer(choices)

                total += 1
                if ca['extra']:
                    commented += 1
                    answer_transformations[ca['text']] = ca['extra'].get('comment')
                else:
                    answer = {
                        'text': ca['text'],
                        'start': ca['start'],
                        'end': ca['end']
                    }
                    yield question, answer, ca['context']


class SquadLoader:
    @staticmethod
    def collapse_spans(answers):
        answers = sorted(answers, key=lambda x: x['start'])

        N = len(answers)
        checked, collapsed = [False] * N, [False] * N         
        for i in range(1, len(answers)):
            if answers[i-1]['question'] == answers[i]['question'] and answers[i-1]['end'] >= answers[i]['start']:
                # overlap
                N1 = answers[i-1]['end'] - answers[i-1]['start']
                N2 = answers[i]['end'] - answers[i]['start']
                if N1 > N2:
                    collapsed[i-1] = True
                else:
                    collapsed[i] = True
                checked[i-1] = True
                checked[i] = True
            else:
                # no overlap
                if not checked[i-1]:
                    collapsed[i-1] = True
                    checked[i-1] = True
        if not checked[N-1]:
            collapsed[N-1] = True
        return [a for i, a in enumerate(answers) if collapsed[i]]

    @staticmethod
    def from_files(fnames, lang='en'):
        total, total_one_answer = 0, 0
        for fname in fnames:
            info = json.load(open(fname))
            if type(info) == dict:
                assert 'data' in info, "Not compatible with SQuAD format"
                data = info['data']
            elif type(info) == list:
                data = info

            for dp in data:
                for paragraph in dp['paragraphs']:
                    context = paragraph['context']
                    answers = []
                    for qa in paragraph['qas']:
                        if ('is_impossible' not in qa) or (not qa['is_impossible']):# and len(qa['answers']) == 1:
                            total_one_answer += 1
                            answer = qa['answers'][0]
                            answer['question'] = qa['question']
                            answer['start'] = answer['answer_start']
                            answer['end'] = answer['answer_start'] + len(answer['text'])
                            del answer['answer_start']
                            if answer not in answers:
                                answers.append(answer)
                        total += 1

                    if answers:
                        for answer in SquadLoader.collapse_spans(answers):
                            yield answer['question'], answer, context
        print(total, total_one_answer)


class TyDiQaLoader:
    @staticmethod
    def from_files(fnames, lang='en'):
        def pretty_stats(dct):
            return "\n".join(["\t{}: {}({}%)".format(k, sdct['v'], round(sdct['v'] * 100 / sdct['t'], 2)) for k, sdct in dct.items()])

        lang2full = {
            'fi': 'finnish',
            'ru': 'russian',
            'en': 'english',
            'ja': 'japanese',
            'te': 'telugu',
            'ar': 'arabic',
            'bn': 'bengali',
            'th': 'thai',
            'id': 'indonesian',
            'ko': 'korean',
            'sw': 'swahili'
        }

        lang = lang2full[lang] # should raise error if language is not present

        total, total_valid, k = 0, 0, 0
        lang_cnt = defaultdict(lambda: {'v': 0, 't': 0})
        for fname in fnames:
            with jsonlines.open(fname) as reader:
                i = 0
                for obj in reader:
                    total += 1
                    lang_cnt[obj['language']]['t'] += 1
                    annotations = obj['annotations']

                    if any([a['yes_no_answer'] != 'NONE' for a in annotations]) or obj['language'] != lang:
                        # skip answers requiring yes or no
                        continue

                    annot_len = len(annotations)
                    if annot_len > 1 and obj['language'] == lang:
                        # all of the languages in the dev set have more than one annotation
                        k += 1

                    doc = obj['document_plaintext'].encode('utf8')
                    candidates = obj['passage_answer_candidates']
                    question = obj['question_text']

                    valid = False
                    for annot in annotations:
                        ca_start = annot['minimal_answer']['plaintext_start_byte']
                        ca_end = annot['minimal_answer']['plaintext_end_byte']
                        passage_index = annot['passage_answer']['candidate_index']

                        if passage_index == -1 or ca_start == 1 or ca_end == 1:
                            continue
                        
                        cand = candidates[passage_index]
                        ca = doc[ca_start:ca_end].decode('utf8', errors="replace")
                        answer = {
                            'text': ca,
                            'start': ca_start - cand['plaintext_start_byte'],
                            'end': ca_end - cand['plaintext_start_byte']
                        }
                        passage = doc[cand['plaintext_start_byte']:cand['plaintext_end_byte']].decode('utf8', errors="replace")
                        total_valid += 1
                        valid = True
                        yield question, answer, passage
                    lang_cnt[obj['language']]['v'] += valid
        # print("Stats:\n{}".format(pretty_stats(lang_cnt)))
        # print("Total valid: {} ({}%)".format(total_valid, round(total_valid / total, 2)))
        # print("More than 1 annotation: {}/{}".format(k, total))


class XquadLoader:
    @staticmethod
    def from_files(fnames, lang='en'):
        total, total_one_answer = 0, 0
        for fname in fnames:
            info = json.load(open(fname))
            assert 'data' in info, "Not compatible with SQuAD format"

            for dp in info['data']:
                for paragraph in dp['paragraphs']:
                    context = paragraph['context']
                    answers = []
                    for qa in paragraph['qas']:
                        if not qa['is_impossible']:# and len(qa['answers']) == 1:
                            total_one_answer += 1
                            answer = qa['answers'][0]
                            answer['question'] = qa['question']
                            answer['start'] = answer['answer_start']
                            answer['end'] = answer['answer_start'] + len(answer['text'])
                            del answer['answer_start']
                            if answer not in answers:
                                answers.append(answer)
                        total += 1

                    if answers:
                        for answer in SquadLoader.collapse_spans(answers):
                            yield answer['question'], answer, context
        print(total, total_one_answer)


import re
import time
import string
import math
from operator import itemgetter
from itertools import product, permutations
from pprint import pprint

import numpy as np

from .common import repr_pos_morph


START_TOKEN = "<START>"
END_TOKEN = "<END>"


def invoke_guards(root, guards_root, return_first=False):
    cand, survivors = [(root.children[0], guards_root.children, [], [])], []
    used_templates = []

    while cand:
        new_cand = []
        for node, clauses, guard_chain, backup in cand:
            if type(clauses) == dict:
                for cname, g in clauses.items():
                    satisfying = g.cond(node, False)#g.parent.is_root())
                    if satisfying:
                        if g.children:
                            for sat in satisfying:
                                gc = list(guard_chain)
                                gc.append(g.rule)
                                new_cand.append((node if guard_chain else sat, g.children, gc, list(g.templates)))
                        elif backup not in used_templates:
                            gc = list(guard_chain)
                            survivors.append((node, backup, gc))
                            used_templates.append(backup)
                        
                        if g.templates and g.templates not in used_templates:
                            gc = list(guard_chain)
                            gc.append(g.rule)
                            survivors.append((node, g.templates, gc))
                            used_templates.append(g.templates)

                            if return_first:
                                return survivors

                    elif backup not in used_templates:
                        gc = list(guard_chain)
                        survivors.append((node, backup, gc))
                        used_templates.append(backup)
        cand = new_cand
    return survivors


def split_coordinate_clauses(tree_root):
    conjuncts = tree_root.select_by("deprel", "conj")
    
    clauses = {}
    for conj in conjuncts:
        clause1_root = conj.parent
        clause2_root = conj

        if clause2_root.child_has_prop("deprel", "nsubj"):
            # means we did find a coordinate clause
            clause1_root.remove_child(clause2_root)
            while clause1_root.deprel != 'root':
                # get to the real root
                clause1_root = clause1_root.parent
            clause2_root.prune('cc')
            clause2_root.make_root()
            clauses[str(clause1_root)] = clause1_root
            clauses[str(clause2_root)] = clause2_root
    return list(clauses.values())


def overgenerate_questions(trees, guards_root, templates, template_examples, return_first=False, distractors=False):
    res = {}
    for root in trees:
        survivors = invoke_guards(root, guards_root, return_first=return_first)

        generated_pairs = set()
        for node, guard_templates, guard_chain in survivors:
            for temp_id in guard_templates:
                temp = templates[temp_id]
                qtemp, atemp, shared_tags = temp['question'], temp['answer'], temp['shared_tags']
                q_expressions, a_expressions = qtemp(node), atemp(node)

                # Example expression elements
                #
                # Q: 
                # [{None: TemplateInstance(None, ['var'])},
                #  {None: TemplateInstance(None, ['ligger'])},
                #  {4: TemplateInstance(node<4>, ['tyngdpunkten'])},
                #  {2: TemplateInstance(node<2>, ['i biomedicinarutbildningen']), 9: TemplateInstance(node<9>, ['till kunskaper om människan och hennes sjukdomar'])}]
                # 
                # A:
                # [{6: TemplateInstance(node<6>, ['på molekylärbiologi'])}
                #  {7: TemplateInstance(node<7>, ['kopplat'])},
                #  {2: TemplateInstance(node<2>, ['i biomedicinarutbildningen']), 9: TemplateInstance(node<9>, ['till kunskaper om människan och hennes sjukdomar'])}]

                # the lists are becoming too large - try to get away with generators as much as possible!
                for temp_question in product(*[q.keys() if q else [] for q in q_expressions]):
                    tq = [x for x in temp_question if x is not None]
                    if len(set(tq)) == len(tq):
                        for temp_answer in product(*[a.keys() if a else [] for a in a_expressions]):
                            ta = [x for x in temp_answer if x is not None]
                            if len(set(ta)) == len(ta):
                                passing = True
                                for s in shared_tags:
                                    q_last, a_last = s.last_in_chain('q'), s.last_in_chain('a')
                                    q = q_expressions[s.q][temp_question[s.q]]
                                    a = a_expressions[s.a][temp_answer[s.a]]

                                    q_check = [q.chain[-1]] if q_last else q.chain
                                    a_check = [a.chain[-1]] if a_last else a.chain

                                    if not (set(q_check) & set(a_check)):
                                        passing = False
                                        break

                                if passing:
                                    for q_seq in product(*[q_expressions[i][k].text for i, k in enumerate(temp_question)]):
                                        for a_seq in product(*[a_expressions[i][k].text for i, k in enumerate(temp_answer)]):
                                            # exclude pronouns for now until pronoun resolution is there
                                            # PUNCT_REGEX = f'[{string.punctuation}]'
                                            # qq = re.sub(PUNCT_REGEX, '', q.lower())
                                            # aa = re.sub(PUNCT_REGEX, '', a.lower())
                                            # if set(qq.split()) & PERSONAL_PRONOUNS or set(aa.split()) & PERSONAL_PRONOUNS:
                                            #     continue

                                            # if any template element was evaluated to an empty sequence
                                            if any([not x.strip() for x in a_seq]) or any([not x.strip() for x in q_seq]):
                                                continue

                                            el = {
                                                'question': [x.strip() for x in q_seq],
                                                'answer': [x.strip() for x in a_seq],
                                                'temp_id': [str(temp_id)],
                                                'guards': [", ".join(guard_chain)],
                                                "base_sentences": [str(root.children[0].get_subtree_text())],
                                            }


                                            if el['answer'] in el['question']:
                                                # Obviously wrong question!
                                                continue

                                            if distractors:
                                                # TODO: fix conv_tree_kernel and then enable again
                                                # dis = generate_distractors(trees, a)
                                                dis = []
                                                el['distractors'] = dis
                                            else:
                                                el['distractors'] = []
                                            
                                            pair = f"{el['question']} => {el['answer']}"
                                            if pair not in res:
                                                res[pair] = el
                                            else:
                                                if el['temp_id'][0] not in res[pair]['temp_id']:
                                                    res[pair]['temp_id'].extend(el['temp_id'])
                                                    res[pair]['guards'].extend(el['guards'])
                                                    res[pair]['base_sentences'].extend(el['base_sentences'])
    return list(res.values())


def get_syntactic_score_old(q_words, unigrams_log_prob, bigrams_log_prob, trigrams_log_prob):
    def backoff(w):
        return w if type(w) == str else w.upos

    lambda4 = 0.000001
    lambda3 = 0.01 - lambda4
    lambda2 = 0.1 - lambda3
    lambda1 = 0.9

    q_words.insert(0, START_TOKEN)
    q_words.append(END_TOKEN)

    N = len(q_words)
    syntactic_score = 0
    if N >= 3:
        N_ngrams = 0
        for i, (w1, w2, w3) in enumerate(zip(q_words[:-2], q_words[1:-1], q_words[2:])):
            w1_token, w2_token, w3_token = repr_pos_morph(w1), repr_pos_morph(w2), repr_pos_morph(w3)
            w1_backoff, w2_backoff, w3_backoff = backoff(w1), backoff(w2), backoff(w3)

            w1u_lprob = unigrams_log_prob.get(w1_token, unigrams_log_prob.get(w1_backoff, float('-inf')))
            
            w1b_lprob = bigrams_log_prob.get(w1_token, bigrams_log_prob.get(w1_backoff, {}))
            w1w2b_lprob = w1b_lprob.get(w2_token, w1b_lprob.get(w2_backoff, float('-inf')))

            w1t_lprob = trigrams_log_prob.get(w1_token, trigrams_log_prob.get(w1_backoff, {}))
            w1w2t_lprob = w1t_lprob.get(w2_token, w1t_lprob.get(w2_backoff, {}))
            w1w2w3t_lprob = w1w2t_lprob.get(w3_token, w1w2t_lprob.get(w3_backoff, float('-inf')))
            N_ngrams += 1

            syntactic_score += np.log(lambda1 * np.exp(w1w2w3t_lprob) + lambda2 * np.exp(w1w2b_lprob) + lambda3 * np.exp(w1u_lprob) + lambda4)
        syntactic_score /= N_ngrams
    elif N == 2:
        w1, w2 = q_words[0], q_words[1]
        w1_token, w2_token = repr_pos_morph(w1), repr_pos_morph(w2)
        w1_backoff, w2_backoff = backoff(w1), backoff(w2)
        w1u_lprob = unigrams_log_prob.get(w1_token, unigrams_log_prob.get(w1_backoff))
        w1b_lprob = bigrams_log_prob.get(w1_token, bigrams_log_prob.get(w1_backoff))
        w1w2b_lprob = w1b_lprob.get(w2_token, w1b_lprob.get(w2_backoff, float('-inf')))
        syntactic_score += np.log(lambda1 * np.exp(w1w2b_lprob) + lambda2 * np.exp(w1u_lprob) + (1 - lambda1 - lambda2))
        syntactic_score /= 2
    elif N == 1:
        w1 = q_words[0]
        w1_token = repr_pos_morph(w1)
        syntactic_score += np.log(lambda1 * np.exp(unigrams_log_prob.get(w1_token, unigrams_log_prob.get(w1.upos, float('-inf')))) + (1 - lambda1))
    return np.exp(syntactic_score)


def get_syntactic_score(q_words, unigrams_log_prob, bigrams_log_prob, trigrams_log_prob):
    def backoff(w):
        return w if type(w) == str else w.upos

    lambda4 = 0.000001
    lambda3 = 0.01 - lambda4
    lambda2 = 0.1 - lambda3
    lambda1 = 0.9

    q_words.insert(0, START_TOKEN)
    q_words.append(END_TOKEN)

    N = len(q_words)
    syntactic_score = 0
    if N >= 3:
        N_ngrams = 0
        for i, w in enumerate(q_words):
            w3_token, w3_backoff = repr_pos_morph(w), backoff(w)
            if i - 1 >= 0:
                w2_token = repr_pos_morph(q_words[i-1])
                w2_backoff = backoff(q_words[i-1])
            else:
                w2_token, w2_backoff = None, None
            
            if i - 2 >= 0:
                w1_token = repr_pos_morph(q_words[i-2])
                w1_backoff = backoff(q_words[i-2])
            else:
                w1_token, w1_backoff = None, None

            w3u_lprob = unigrams_log_prob.get(w3_token, unigrams_log_prob.get(w3_backoff, float('-inf')))
            
            if w2_token:
                w2b_lprob = bigrams_log_prob.get(w2_token, bigrams_log_prob.get(w2_backoff, {}))
                w2w3b_lprob = w2b_lprob.get(w3_token, w2b_lprob.get(w3_backoff, float('-inf')))
            else:
                w2w3b_lprob = float('-inf')

            if w1_token:
                w1t_lprob = trigrams_log_prob.get(w1_token, trigrams_log_prob.get(w1_backoff, {}))
                w1w2t_lprob = w1t_lprob.get(w2_token, w1t_lprob.get(w2_backoff, {}))
                w1w2w3t_lprob = w1w2t_lprob.get(w3_token, w1w2t_lprob.get(w3_backoff, float('-inf')))
            else:
                w1w2w3t_lprob = float('-inf')
            
            N_ngrams += 1

            syntactic_score += np.log(lambda1 * np.exp(w1w2w3t_lprob) + lambda2 * np.exp(w2w3b_lprob) + lambda3 * np.exp(w3u_lprob) + lambda4)
        syntactic_score /= N_ngrams
    elif N == 2:
        w1, w2 = q_words[0], q_words[1]
        w1_token, w2_token = repr_pos_morph(w1), repr_pos_morph(w2)
        w1_backoff, w2_backoff = backoff(w1), backoff(w2)
        w2u_lprob = unigrams_log_prob.get(w2_token, unigrams_log_prob.get(w2_backoff))
        w1b_lprob = bigrams_log_prob.get(w1_token, bigrams_log_prob.get(w1_backoff))
        w1w2b_lprob = w1b_lprob.get(w2_token, w1b_lprob.get(w2_backoff, float('-inf')))
        syntactic_score += np.log(lambda1 * np.exp(w1w2b_lprob) + lambda2 * np.exp(w2u_lprob) + (1 - lambda1 - lambda2))
        syntactic_score /= 2
    elif N == 1:
        w1 = q_words[0]
        w1_token = repr_pos_morph(w1)
        syntactic_score += np.log(lambda1 * np.exp(unigrams_log_prob.get(w1_token, unigrams_log_prob.get(w1.upos, float('-inf')))) + (1 - lambda1))
    return np.exp(syntactic_score)


def rank(res, stanza_pipe, stanza_dep_pipe, qw_stat, a_tmpl, log_prob, rtl=False, join_char=' '):
    """
    Rank the overgenerated questions
    
    :param      res:              The list of overgenerated questions
    :type       res:              list
    :param      stanza_pipe:      Stanza pipeline with processors "tokenize,pos"
    :type       stanza_pipe:      stanza.Pipeline
    :param      stanza_dep_pipe:  Stanza pipeline with processors "tokenize,pos,lemma,depparse"
    :type       stanza_dep_pipe:  stanza.Pipeline
    """
    unigrams_log_prob, bigrams_log_prob, trigrams_log_prob = log_prob
    generated_templates = set()
    atemp_freq, qw_freq = [], []
    q_synt, q_sem, qw_freq, atemp_freq = [], [], [], []
    for cand in res:
        # each cand contains question and answer divided into tokens as in the template
        # but those might not necessarily be correct for languages like Japanese or Chinese
        q = join_char.join(cand['question']).strip()
        a = join_char.join(cand['answer']).strip()
        if not q or not a:
            # to preserve indices in `synt` as in `res`
            q_synt.append(float('-inf'))
            qw_freq.append(0)
            atemp_freq.append(0)
            continue

        q_words = stanza_pipe(q).sentences[0].words
        a_words = stanza_dep_pipe(a).sentences[0].words
        aw = None
        for i, w in enumerate(a_words):
            if w.deprel == 'root':
                aw = i
                break
        if aw is None:
            q_synt.append(float('-inf'))
            qw_freq.append(0)
            atemp_freq.append(0)
            continue

        a_pos_pattern = [repr_pos_morph(x) for x in a_words]
        
        atemp_freq.append(a_tmpl[" ".join(a_pos_pattern)])
        qw_key = (q_words[-1] if rtl else q_words[0]).text.lower()

        if '/' in a_pos_pattern[aw]:
            a_pos, a_morph = a_pos_pattern[aw].split('/')
            a_morph = set(a_morph.split('|'))
        else:
            a_pos, a_morph = a_pos_pattern[aw], None

        if qw_key not in qw_stat:
            if rtl:
                cand['question'] = " ".join([x.text for x in q_words[:-1]]) + qw_key
            else:
                cand['question'] = qw_key + " " + " ".join([x.text for x in q_words[1:]])

        valid_freq, total_freq = 0, 0
        for pos_pattern, freq in qw_stat[qw_key].items():
            if '/' in pos_pattern:
                pos, morph = pos_pattern.split('/')
                morph = morph.split('|')
                if a_pos == pos and a_morph and a_morph & set(morph) == a_morph:
                    valid_freq += freq
            else:
                if a_pos == pos_pattern:
                    valid_freq += freq
            total_freq += freq

        if total_freq > 0:
            qw_freq.append(round(valid_freq / total_freq, 4))
        else:
            qw_freq.append(0)

        q_synt.append(round(get_syntactic_score(q_words, unigrams_log_prob, bigrams_log_prob, trigrams_log_prob), 4))

    scores = [0.8 * q_synt[x] + 0.2 * qw_freq[x] for x in range(len(res))]

    idx_sorted_by_scores = sorted(
        list(range(len(res))),
        key=lambda x: scores[x],
        reverse=True
    )
    return idx_sorted_by_scores, qw_freq, atemp_freq, scores


"""
Common utilities for Stanza resources.
"""

import os
import glob
import requests
from pathlib import Path
import json
import hashlib
import zipfile
import shutil

import dill
from tqdm.auto import tqdm

from .common import (
    DEFAULT_TEMPLATES_DIR, QUINDUCTOR_RESOURCES_GITHUB, MODELS,
    get_logger, get_default_model_path, load_pos_ngrams
)
from .guards import load_guards
from .rules import load_templates, load_template_examples


logger = get_logger()


# The functions below (until -- END STANZA --) are a snapshot taken from Stanza
# https://github.com/stanfordnlp/stanza/blob/f91ca215e175d4f7b202259fe789374db7829395/stanza/resources/common.py

def ensure_dir(path):
    """
    Create dir in case it does not exist.
    """
    Path(path).mkdir(parents=True, exist_ok=True)

def get_md5(path):
    """
    Get the MD5 value of a path.
    """
    with open(path, 'rb') as fin:
        data = fin.read()
    return hashlib.md5(data).hexdigest()

def unzip(path, filename):
    """
    Fully unzip a file `filename` that's in a directory `dir`.
    """
    logger.debug(f'Unzip: {path}/{filename}...')
    with zipfile.ZipFile(os.path.join(path, filename)) as f:
        f.extractall(path)

def get_root_from_zipfile(filename):
    """
    Get the root directory from a archived zip file.
    """
    zf = zipfile.ZipFile(filename, "r")
    assert len(zf.filelist) > 0, \
        f"Zip file at f{filename} seems to be corrupted. Please check it."
    return os.path.dirname(zf.filelist[0].filename)

def file_exists(path, md5):
    """
    Check if the file at `path` exists and match the provided md5 value.
    """
    return os.path.exists(path) and get_md5(path) == md5

def assert_file_exists(path, md5=None):
    assert os.path.exists(path), "Could not find file at %s" % path
    if md5:
        file_md5 = get_md5(path)
        assert file_md5 == md5, "md5 for %s is %s, expected %s" % (path, file_md5, md5)

def download_file(url, path, proxies, raise_for_status=False):
    """
    Download a URL into a file as specified by `path`.
    """
    verbose = logger.level in [0, 10, 20]
    r = requests.get(url, stream=True, proxies=proxies)
    with open(path, 'wb') as f:
        file_size = int(r.headers.get('content-length'))
        default_chunk_size = 131072
        desc = 'Downloading ' + url
        with tqdm(total=file_size, unit='B', unit_scale=True, \
            disable=not verbose, desc=desc) as pbar:
            for chunk in r.iter_content(chunk_size=default_chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    pbar.update(len(chunk))
    if raise_for_status:
        r.raise_for_status()
    return r.status_code

def request_file(url, path, proxies=None, md5=None, raise_for_status=False):
    """
    A complete wrapper over download_file() that also make sure the directory of
    `path` exists, and that a file matching the md5 value does not exist.
    """
    ensure_dir(Path(path).parent)
    if file_exists(path, md5):
        logger.info(f'File exists: {path}.')
        return
    download_file(url, path, proxies, raise_for_status)
    assert_file_exists(path, md5)

# -- END STANZA --


def download(lang):
    # verify if the model is already downloaded and skip if it is (resume if downloaded partially)
    if lang in MODELS:
        logger.info("Downloading Quinductor templates for {}".format(lang))
        lang_dir = os.path.join(DEFAULT_TEMPLATES_DIR, lang)
        for fname in ['atmpl.dill', 'qwstats.dill', 'idf_{}.csv'.format(lang)]:
            request_file(
                '{}/{}/{}'.format(QUINDUCTOR_RESOURCES_GITHUB, lang, fname),
                os.path.join(lang_dir, fname)
            )
        
        pos_ngrams_dir = os.path.join(lang_dir, 'pos_ngrams')
        for fname in MODELS[lang]['pos_ngrams']:
            request_file(
                '{}/{}/{}/{}'.format(QUINDUCTOR_RESOURCES_GITHUB, lang, 'pos_ngrams', fname),
                os.path.join(pos_ngrams_dir, fname)
            )

        model_dir = os.path.join(lang_dir, str(MODELS[lang]['templates']))
        for fname in ['guards.txt', 'templates.txt', 'sentences.txt']:
            request_file(
                '{}/{}/{}/{}'.format(QUINDUCTOR_RESOURCES_GITHUB, lang, str(MODELS[lang]['templates']), fname),
                os.path.join(model_dir, fname)
            )
        logger.info("Finished downloading Quinductor templates (saved to {})".format(lang_dir))
    else:
        logger.warning('Templates for language {} are not available.'.format(lang))



def use(lang):
    templates_folder = get_default_model_path(lang)

    ranking_folder = Path(templates_folder).parent
    ng_folder = os.path.join(ranking_folder, 'pos_ngrams')

    return {
        'pos_ngrams': load_pos_ngrams(ng_folder),
        'guards': load_guards(glob.glob(os.path.join(templates_folder, 'guards.txt'))),
        'templates': load_templates(glob.glob(os.path.join(templates_folder, 'templates.txt'))),
        'examples': load_template_examples(glob.glob(os.path.join(templates_folder, 'sentences.txt'))),
        'qw_stat': dill.load(open(os.path.join(ranking_folder, 'qwstats.dill'), 'rb')),
        'a_stat': dill.load(open(os.path.join(ranking_folder, 'atmpl.dill'), 'rb'))
    }
    


# -*- coding: utf-8 -*-
import json
import dill
import argparse
import os
import math
import sys
import time
import string
import shutil
import re
import itertools
from collections import defaultdict
from operator import itemgetter
from difflib import SequenceMatcher

import tqdm

import stanza
from stanza.utils.conll import CoNLL

import udon2

from .loaders import *
from .repro import get_client
from .common import *
from .guards import *
from .rules import load_templates


logger = get_logger()


def get_negatives(node, chain):
    # get all possible negatives, which is all children of the node and the node itself if it has any children
    negatives = [chain + [f"{c.deprel}#{int(c.id)}"] for c in node.children]
    if chain and len(node.children) > 0:
        node_el = list(chain)
        node_el[-1] += "*"
        negatives.append(node_el)
    return negatives


def shift_reduce(template_elements):
    template, children_map = [], {}
    N = len(template_elements)
    for t in template_elements:
        logger.debug(" ".join(map(str, template)) + " <-- " + str(t))
        if template:
            logger.debug("type(template[-1]): {}".format(type(template[-1])))
            if type(template[-1]) == TemplateElement:
                logger.debug("\ttemplate[-1] is lemma? -- {}".format(template[-1].is_lemma))
        logger.debug("type(t): {}".format(type(t)))
        if type(t) == TemplateElement:
            logger.debug("\tt is lemma? -- {}".format(t.is_lemma))
        if template and type(template[-1]) == TemplateElement and type(t) == TemplateElement and not template[-1].is_lemma and not t.is_lemma:
            logger.debug("stack contains at least two items!")
            intersection = template[-1] & t
            N1, N2, Ni = len(template[-1]), len(t), len(intersection)
            logger.debug("intersection: {}".format(str(intersection)))
            logger.debug("N1: {}, N2: {}, Ni: {}".format(N1, N2, Ni))

            if intersection and (N1 - Ni <= 1 or N2 - Ni <= 1):
                # If the intersection exists and it's one level above any of the two candidates,
                # we can collapse them then. The restriction on the levels is needed
                # to avoid huge collapses that will probably be non-generalizable, i.e.
                # <w.conj - obl5*>, whereas the conj structures can vary a lot and hence
                # some wrong questions will be generated
                
                logger.debug("==> REDUCE")

                negatives = []
                for node in [t, template[-1]]:
                    new_node, new_chain = node.node, node.chain
                    new_chain_str = ".".join(new_chain)
                    if new_chain_str not in children_map:
                        new_neg = get_negatives(new_node, new_chain)
                        children_map[new_chain_str] = new_neg
                    negatives.extend(children_map[new_chain_str])

                    while new_chain != intersection:
                        new_node = new_node.parent
                        new_chain = new_chain[:-1]

                        new_chain_str = ".".join(new_chain)
                        if new_chain_str not in children_map:
                            new_neg = get_negatives(new_node, new_chain)
                            children_map[new_chain_str] = new_neg
                        negatives.extend(children_map[new_chain_str])

                logger.debug("\tnegatives: {}".format(negatives))

                # means they are part of the same subtree
                te = TemplateElement(intersection, new_node, t.id, True)
                for n in [template[-1], t]:
                    if len(n.node.children) > 0 and not n.is_subtree:
                        n.chain[-1] += "*"

                te.copy_positives([template[-1], t]) # copy only positives?
                te.add_positives([template[-1].chain, t.chain]) # if those are copied, then this addition is not necessary?
                te.copy_negatives([template[-1], t])
                te.add_negatives(negatives)

                logger.debug("\treduced token: {}".format(str(te)))

                template.pop()
                template.append(te)
            else:
                logger.debug("==> SHIFT: the token {} is added to stack".format(str(t)))
                # means they are from different subtrees
                template.append(t)
        else:
            logger.debug("==> SHIFT: the token {} is added to stack".format(str(t)))
            # just the second word
            template.append(t)
    return template, children_map


def generate_question_template(s_root, q_root, strict=True, join_char=' '):
    """
    Generates question template by first linearizing the question dep parse tree
    and then trying to get a question word from the beginning nodes of the linear order.

    If the node `n` is the question root itself, add a node template element [w]
     
    :param      s_root:  The node representing the root word of the original sentence
    :type       s_root:  udon2.Node
    :param      q_root:  The node representing the root word of the question
    :type       q_root:  udon2.Node
    """
    node_list = q_root.linear()
    N = len(node_list)

    logger.debug("-- S-TRANSFORM --")
    template_elements, is_template_element = [], []
    for i in range(N):
        # can't use `node_list[i].get_rel() != "punct"`, since sometimes adpositions become punct!
        if node_list[i].form not in string.punctuation:
            if s_root.form.lower() == node_list[i].form.lower():
                # means it's a root of the sentence tree itself
                template_elements.append([TemplateElement([], s_root)])
                is_template_element.append(True)
            elif s_root.lemma == node_list[i].form.lower():
                template_elements.append([TemplateElement([], s_root, is_lemma=True)])
                is_template_element.append(True)
            else:
                # dependency relation can be different and POS-tag as well (e.g. "har" can be both AUX and VERB)
                # morphological properties should typically be the same though
                s_subtree = s_root.select_by("form", node_list[i].form) # by lemma and text
                if len(s_subtree) == 0:
                    s_subtree = s_root.select_by("form", node_list[i].form.lower())
                
                if len(s_subtree) == 0:
                    # try to match by lemma for the cases of "When does he play football?" and the sentence "He plays football every week."
                    s_lemma = s_root.select_by("lemma", node_list[i].form.lower())

                    if len(s_lemma) == 0:
                        if not strict:
                            template_elements.append([node_list[i].form])
                            is_template_element.append(False)
                        else:
                            # print("0-length!")
                            # means we can't find one of the words!
                            return None, False, False
                    else:
                        candidates = []
                        for cand in s_lemma:
                            root_chain = chain(s_root, cand, include_ids=True)
                            candidates.append(TemplateElement(root_chain, cand, is_lemma=True))

                        template_elements.append(candidates)
                        is_template_element.append(True)
                elif len(s_subtree) == 1:
                    cand = s_subtree[0]
                    root_chain = chain(s_root, cand, include_ids=True)
                    template_elements.append([TemplateElement(root_chain, cand)])
                    is_template_element.append(True)
                else:
                    candidates = []
                    for cand in s_subtree:
                        root_chain = chain(s_root, cand, include_ids=True)
                        candidates.append(TemplateElement(root_chain, cand))

                    if not candidates:
                        # print("0-cand")
                        # we don't want to work with those not connected to question phrase
                        # where we can't find the candidate connected to the parent
                        return None, False, False

                    template_elements.append(candidates)
                    is_template_element.append(True)

    min_diff, best_cand = float('inf'), None

    pair_check = list(zip(is_template_element[:-1], is_template_element[1:]))

    cand_processed, cand_limit = 0, 200 # computational simplification for processing potentially long sentences
    # problem with the number of hops: how would you calculate the number of hops if they are on the different side of the root?
    # sum before the root + sum after the root? not exactly clear
    for cand in itertools.product(*template_elements):
        if cand_processed >= cand_limit: break
        if len(set(cand)) == len(cand):
            diff = sum([abs(cand[i+1].id - cand[i].id) for i, (n1, n2) in enumerate(pair_check)
                if n1 and n2 and cand[i].id is not None and cand[i+1].id is not None])
            if diff < min_diff:
                min_diff = diff
                best_cand = cand
        cand_processed += 1

    if best_cand is None and min_diff == float('inf'):
        best_cand = cand

    logger.debug(list(map(str, best_cand)))
    
    logger.debug("-- SHIFT-REDUCE --")

    template, children_map = shift_reduce(best_cand)

    logger.debug(" ".join(map(str, template)))

    if len(template) == 1:
        logger.debug("-- OVERGENERIC! RETURN S-TRANSFORM --")
        # overgeneric case, like what <w.obl - case - num>?
        # simply go on with an original word by word template after S-transform
        best_cand = list(best_cand)
        S_t = sum([type(x) == TemplateElement for x in best_cand])
        return best_cand

    logger.debug("-- MERGING NEGATIVES --")
    for t in template:
        if type(t) == TemplateElement:
            t.merge_negatives(children_map)

    logger.debug(join_char.join(map(str, template)))
    logger.debug("-- END OF TRANSFORMATION --\n")
    return template


def generate_answer_template(s_root_word, answer_str, join_char=' '):
    # try to find the right subtree containing the answer_str
    answer_node = s_root_word.textual_intersect(answer_str.strip())
    logger.debug(f"-- ANSWER STRING -> {answer_str} --")
    
    if answer_node:
        logger.debug(f"-- FOUND ANSWER NODE -> {answer_node.form} --")
        logger.debug(f"-- SUBTREE ANSWER NODE -> {answer_node.get_subtree_text()} --")
    else:
        # this case should not happen normally for Swedish, because we know for sure
        # the answer is in the original sentence - that's how the dataset is structured
        # but for other datasets it's definitely a possibility
        logger.debug("-- ANSWER NOT PRESENT --\n")
        return None

    # find the way from the answer node of the original sentence to its root
    if answer_node.form.lower() == answer_str.lower():
        root_chain = chain(s_root_word, answer_node, include_ids=True)
        template = [TemplateElement(root_chain, answer_node)]
        logger.debug(" ".join(map(str, template)))
        logger.debug("-- END OF TRANSFORMATION --\n")
    elif answer_node.get_subtree_text().lower() == answer_str.lower():
        root_chain = chain(s_root_word, answer_node, include_ids=True)
        template = [TemplateElement(root_chain, answer_node, is_subtree=True)]
        logger.debug(" ".join(map(str, template)))
        logger.debug("-- END OF TRANSFORMATION --\n")
    else:
        sketch, start = [], None
        for i, word in enumerate(re.findall(r"[\w']+|[.,!?;]", answer_str.strip())):
            res = answer_node.select_by("form", word)
            if len(res) == 1:
                if not start: start = i
                root_chain = chain(s_root_word, res[0], include_ids=True)
                sketch.append(TemplateElement(root_chain, res[0]))
            elif res:
                res2 = []
                for r in res:
                    root_chain = chain(s_root_word, r, include_ids=True)
                    res2.append(TemplateElement(root_chain, r))
                sketch.append(res2)
            else:
                logger.debug("-- COULDN'T FIND: res, word --")
                logger.debug("-- NO ANSWER COULD BE GENERATED --\n")
                return None

        logger.debug(f"-- START -> {start} --")

        if not start:
            logger.debug("-- NO ANSWER COULD BE GENERATED --\n")
            return None

        # we know for sure that at least one word will match exactly
        for j in list(range(0, start)) + list(range(start + 1, i + 1)):
            if type(sketch[j]) == list:
                min_dist, best_match = float('inf'), None
                for el in sketch[j]:
                    if j > 0 and el.id - sketch[j-1].id == 1:
                        # case of the exact following
                        best_match = el
                        break

                    d = abs(el.id - sketch[start].id)
                    if d < min_dist:
                        min_dist = d
                        best_match = el
                sketch[j] = best_match
        
        logger.debug(f"ANSWER SKETCH --- {list(map(str, sketch))}")

        template, children_map = shift_reduce(sketch)

        for t in template:
            if type(t) == TemplateElement:
                t.merge_negatives(children_map)
        
        logger.debug(join_char.join(map(str, template)))
        logger.debug("-- END OF TRANSFORMATION --\n")

    return template


def process_modifiers(templates):
    """
    If there are any modifiers (advmod or amod), generate identical templates, but without
    these modifiers
    
    :param      templates:  The templates
    :type       templates:  dict
    """

    to_process = list(templates.items())
    search_temp = [f"<w.{x}>" for x in MODIFIER_RELS]
    for temp, sent in to_process:
        tokens = temp.split()
        new_tokens = []
        answer_started = False
        for token in tokens:
            if token == "=>":
                answer_started = True
                new_tokens.append(token)
            elif token not in search_temp or answer_started:
                new_tokens.append(token)
        templates[" ".join(new_tokens)] = sent
    return templates


def replace_root_by_chain(temp, chain_str):
    return temp.replace("w.", f"w.{chain_str}.")\
               .replace('<w>', f"<w.{chain_str}>")\
               .replace('[w]', f"[w.{chain_str}]")


def normalize_templates(qtemp, atemp):
    def __norm_temp(x):
        nonlocal idx, mapping
        for i in range(len(x)):
            if type(x[i]) == TemplateElement and x[i].chain:
                new_chain = []
                for el in x[i].chain:
                    if el not in mapping:
                        mapping[el] = f"{el.split('#')[0]}#{idx}"
                        idx += 1
                    new_chain.append(mapping[el])

                positives = []
                kept = set()
                for lst in x[i].positives:
                    new_lst = []
                    for j in range(len(lst)):
                        exact = '*' in lst[j]
                        lst_j = lst[j].replace('*', '')
                        if lst_j in mapping:
                            new_lst.append(mapping[lst_j] + ('*' if exact else ''))
                        else:
                            kept.add(lst[j])
                            new_lst.append(lst[j])
                    positives.append(new_lst)

                negatives = []
                for lst in x[i].negatives:
                    new_lst = []
                    for j in range(len(lst)):
                        if lst[j] in kept:
                            new_lst.append(lst[j])
                        else:
                            exact = '*' in lst[j]
                            lst_j = lst[j].replace('*', '')
                            if lst_j not in mapping:
                                mapping[lst_j] = f"{lst_j.split('#')[0]}#{idx}"
                                idx += 1

                            new_lst.append(mapping[lst_j] + ('*' if exact else ''))
                    negatives.append(new_lst)

                new_el = TemplateElement(new_chain, x[i].node, last_span_id=x[i].id, is_subtree=x[i].is_subtree, is_lemma=x[i].is_lemma)
                new_el.add_positives(positives)
                new_el.add_negatives(negatives)

                x[i] = new_el
        return x

    idx = 1
    mapping = {}
    return __norm_temp(qtemp), __norm_temp(atemp)


def generate_templates(fname, stanza_lang, rtl=False, min_support=2, strict=True, case_folding=False, remove_punct=False,
                       temp_fname="gen_templates.txt", sent_fname="gen_sentences.txt", remove_diacritics=True, dot_fix=False,
                       join_char=' ', idf_file=None):
    def record(key):
        problems[key] += 1

    def print_report(templates):
        print("{} templates".format(len(templates)))
        print("{} impossible questions".format(problems['impossible']))
        print("{} possible questions of which {} share the root with the original sentence".format(problems['possible'], problems['same_root']))
        print("{} copula questions".format(problems['copula']))
        print("{} have no question templates".format(problems['no_q_template']))
        print("{} have no answer templates".format(problems['no_a_template']))

    problems = {
        'possible': 0,
        'impossible': 0,
        'same_root': 0,
        'copula': 0,
        'no_q_template': 0,
        'no_a_template': 0
    }

    templates = {}
    with open(fname) as f:
        for line in tqdm.tqdm(f):
            if line.strip():
                question, answer, base_sentence = line.split(" #|@ ")

                if remove_diacritics:
                    question = remove_unicode_diacritics(question)
                    base_sentence = remove_unicode_diacritics(base_sentence)
                    answer = remove_unicode_diacritics(answer)

                if case_folding:
                    question, answer, base_sentence = question.lower(), answer.lower(), base_sentence.lower()

                # lowercasing is a necessary step to mitigate parser's errors
                if remove_punct:
                    question = remove_unicode_punctuation(question)
                    base_sentence = remove_unicode_punctuation(base_sentence)
                    answer = remove_unicode_punctuation(answer)

                question, base_sentence, answer = question.strip(), base_sentence.strip(), answer.strip()
                if dot_fix:
                    if not is_punctuation(question[-1]):
                        question += "?"
                    if not is_punctuation(base_sentence[-1]):
                        base_sentence += "."

                # have to proceed through files, because C++ package works with files
                with open('sentence.conll', 'w') as f1:
                    ss = stanza_lang(base_sentence)
                    conll_list = CoNLL.convert_dict(ss.to_dict())
                    sentence_tokenized = [w.text for s in ss.sentences for w in s.words]
                    f1.write(CoNLL.conll_as_string(conll_list))

                with open('question.conll', 'w') as f1:
                    qq = stanza_lang(question)
                    conll_list = CoNLL.convert_dict(qq.to_dict())
                    question_tokenized = [w.text for s in qq.sentences for w in s.words]
                    f1.write(CoNLL.conll_as_string(conll_list))

                ud_s = udon2.ConllReader.read_file('sentence.conll')[0]
                ud_q = udon2.ConllReader.read_file('question.conll')[0]
                
                # s_roots = udon2.ConllReader.read_file('sentence.conll')
                # q_roots = udon2.ConllReader.read_file('question.conll')
                # ud_s = s_roots[0]
                # ud_q = q_roots[0]

                s_root_word = ud_s.children[0]
                q_root_word = ud_q.children[0]

                if strict:
                    diff = get_difference(question_tokenized, sentence_tokenized)
                    cond = not diff
                else:
                    same = get_intersection(question_tokenized, sentence_tokenized)
                    cond = len(same) > 0
                
                if cond:
                    # means there's a direct dependency tree transformation!
                    record('possible')
                    if s_root_word.form.lower() == q_root_word.form.lower():
                        # many questions that can be asked share the root with a sentence
                        record('same_root')
                    elif q_root_word.prop_exists("deprel", "cop"):
                        # means this is a copula question
                        record('copula')

                    q_temp = generate_question_template(s_root_word, q_root_word, strict=strict, join_char=join_char)

                    to_check = q_temp[:-1] if rtl else q_temp[1:]
                    S_t = sum([type(x) == TemplateElement for x in to_check])
                    S_nt = len(to_check) - S_t

                    if not q_temp:
                        record('no_q_template')
                        continue
                    if S_t == 0:
                        continue
                    
                    qw = q_temp[-1] if rtl else q_temp[0]
                    if type(qw) == TemplateElement:
                        # the first word is not a constant, so no question word there
                        continue
                    
                    if rtl:
                        qw = q_temp.pop()
                        q_temp.append('<qw>')
                    else:
                        qw = q_temp.pop(0)
                        q_temp.insert(0, '<qw>')

                    a_temp = generate_answer_template(s_root_word, answer, join_char=join_char)
                    if not a_temp:
                        record('no_a_template')
                        continue

                    q_temp, a_temp = normalize_templates(q_temp, a_temp)
                    qtemp_without_qw = join_char.join(map(str, q_temp))

                    if qtemp_without_qw not in templates:
                        templates[qtemp_without_qw] = {
                            'question': q_temp,
                            'all_templates': S_nt == 0, 
                            'answer': a_temp,
                            'qw': {}
                        }

                    assert templates[qtemp_without_qw]['all_templates'] == (S_nt == 0), "Inconsistency in templates found"

                    if qw not in templates[qtemp_without_qw]['qw']:
                        templates[qtemp_without_qw]['qw'][qw] = {}

                    atemp_str = join_char.join(map(str, a_temp))
                    if atemp_str not in templates[qtemp_without_qw]['qw'][qw]:
                        templates[qtemp_without_qw]['qw'][qw][atemp_str] = {
                            'answer': a_temp,
                            'examples': []
                        }

                    templates[qtemp_without_qw]['qw'][qw][atemp_str]['examples'].append({
                        'sentence': base_sentence.strip(),
                        'question': question.strip(),
                        'answer': answer.strip(),
                        'node': s_root_word.copy(), # If not copying, then we'll have a memory error, since the associated TreeList will be freed
                    })

                    # templates[f"{s_root_word.upos} #|@ {str(s_root_word.feats)} #|@ {s_root_word.child_has_prop('deprel', 'aux')} #|@ {non_temp_el} #|@ {q_temp}"][qw][a_temp].add(
                    #     base_sentence.strip() + " | " + question.strip() + " | " + answer.strip())
                elif strict:
                    record('impossible')

    idf = load_idf(idf_file) if idf_file else None

    final_templates, temp_id = [], 1
    temp_base = os.path.splitext(os.path.basename(temp_fname))[0]
    with open(temp_fname, "w") as f, open(sent_fname, 'w') as f1:
        for _, passport in templates.items():
            N_ex = sum([len(data['examples']) for _, endings in passport['qw'].items() for _, data in endings.items()])

            q_tmpl = join_char.join(map(str, passport['question']))

            if passport['all_templates'] or N_ex >= min_support:
                idfs = [idf.get(t, float('inf')) for t in passport['question'] if type(t) == str and t != '<qw>']
                max_idf = max(idfs) if idfs else 0
                if max_idf <= math.log(4): # appeared in at least 25% of the documents
                    for qw, endings in passport['qw'].items():
                        for a_tmpl, data in endings.items():
                            logger.debug("-- {} - {} - {} -> PASSED --".format(q_tmpl, passport['all_templates'], N_ex))

                            final_templates.append({
                                'question': q_tmpl.replace("<qw>", qw),
                                'answer': a_tmpl,
                                'props': [
                                    {
                                        'pos': x['node'].upos,
                                        'has_aux': x['node'].child_has_prop('deprel', 'aux'),
                                        'feats': x['node'].feats
                                    } for x in data['examples']
                                ]
                            })

                            sent = "\n".join([" | ".join([x['sentence'], x['question'], x['answer']]) for x in data['examples']])

                            tmpl = "{} => {}".format(q_tmpl.replace('<qw>', qw), a_tmpl)
                            f.write("{}\n".format(tmpl))
                            f1.write("id: {}{}\n{}\n\n".format(temp_base, temp_id, sent))
                            temp_id += 1
                else:
                    logger.debug("-- {} - {} - {} -> FAILED IDF ({}) --".format(q_tmpl, passport['all_templates'], N_ex, max_idf))
            else:
                logger.debug("-- {} - {} - {} -> FAILED --".format(q_tmpl, passport['all_templates'], N_ex))

    print_report(final_templates)

    return final_templates, temp_fname


def write_answer_transformations(transformations, fname):
    with open(fname, 'w') as f:
        for i, (ca, comment) in enumerate(transformations.items()):
            f.write(f"#CASE{i+1}\n")
            f.write(f"CA\t{ca}\n")
            f.write(f"COM\t{comment}\n")
            f.write("\n")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--lang', type=str, help='A language for template generation (en, sv are currently supported)')
    parser.add_argument('-d', '--data', type=str, help='Comma-separated list of files to generate questions from')
    parser.add_argument('-ft', '--format', type=str, help='Data format (tt for Textinator or squad for Squad)')
    parser.add_argument('-k', '--min-support', default=1, type=int, help="""
        A minimum number of sentences (excluding the question word) required for the template to be included.
        Applied only to the templates with at least one constant expression (excluding the first element, which
        is assumed to be the question word).""")
    parser.add_argument('-st', '--strict', action='store_true') 
    parser.add_argument('-f', '--force', action='store_true', help='Re-read data and ignore cache')
    parser.add_argument('-t', '--templates', type=str, default='NA', help='A template file to generate guards for')
    parser.add_argument('-cf', '--case-folding', action='store_true', help='Lowercase everything if true')
    parser.add_argument('-rp', '--remove-punct', action='store_true', help='Remove punctuation if true')
    parser.add_argument('-mdb', '--modeldb-url', type=str, default='')
    parser.add_argument('-tdir', '--template-dir', type=str, default='tmpl', help='A directory for storing templates')
    parser.add_argument('-rtl', '--right-to-left', action='store_true')
    parser.add_argument('-rd', '--remove-diacritics', action='store_true')
    parser.add_argument('-idf', '--idf', type=str, default='')
    parser.add_argument('--dot-fix', action='store_true', help="""
        Append dots at the end of the sentences even if punctuation is removed in attempt to fix dependency trees""")
    parser.add_argument('-j', '--join-char', type=str, default=" ")
    parser.add_argument('-qac', '--qac-file', type=str, help='A preprocessed .qac file for the given dataset')
    args = parser.parse_args()

    if not args.idf:
        default_file = os.path.join(DEFAULT_TEMPLATES_DIR, args.lang, 'idf_{}.csv'.format(args.lang))
        if os.path.exists(default_file):
            args.idf = default_file
        else:
            logger.error(
                """No valid IDF file was provided. Either download it by running `quinductor.download({})`
                or create your own and supply it using `-idf` argument""".format(args.lang)
            )
            sys.exit(1)

    # arabic, finnish - include mwt
    # russian - exclude mwt
    dep_proc = 'tokenize,lemma,mwt,pos,depparse' if args.lang in ['fi', 'ar'] else 'tokenize,lemma,pos,depparse'
    proc = 'tokenize,mwt,pos' if args.lang in ['fi', 'ar'] else 'tokenize,pos'

    start = time.time()
    timestamp = str(start).replace('.', '')

    if args.modeldb_url:
        client = get_client(args.modeldb_url)
        client.set_project("SQuAD")#"TyDiQA QG")#"SvQG")
        client.set_experiment(args.lang)
        run = client.set_experiment_run("{}_{}".format(args.lang, timestamp))
        join_char = args.join_char
        if join_char == ' ':
            join_char = '<space>'
        elif join_char == '':
            join_char = '<nothing>'

        run.log_hyperparameters({
            'min_support': args.min_support,
            'data': args.data,
            'strict': args.strict,
            'case_folding': args.case_folding,
            'remove_punct': args.remove_punct,
            'right_to_left': args.right_to_left,
            'stanza_processors': dep_proc,
            'dot_fix': args.dot_fix,
            'min_support': args.min_support,
            'remove_diacritics': args.remove_diacritics,
            'join_character': join_char,
            'idf': args.idf
        })

    questions_fnames = [x.strip() for x in args.data.split(',')]

    DIR = os.path.join(args.template_dir, args.lang, timestamp)
    if not os.path.exists(DIR):
        os.makedirs(DIR)

    gen_fname = os.path.join(DIR, f'{args.lang}.qac')

    if not os.path.exists(gen_fname) or args.force:
        if args.qac_file:
            shutil.copy(args.qac_file, gen_fname)
        else:
            stanza_tokenizer = stanza.Pipeline(lang=args.lang, processors=proc)

            if args.format == 'tt':
                data_loader = TextinatorLoader
            elif args.format == 'squad':
                data_loader = SquadLoader
            elif args.format == 'tydiqa':
                data_loader = TyDiQaLoader
            else:
                print("Unrecognized data format")
                sys.exit(1)

            with open(gen_fname, 'w') as f:
                for q, a, c in data_loader.from_files(questions_fnames, args.lang):
                    sent = get_sentence_by_answer(a, c, stanza_tokenizer)
                    if sent:
                        f.write(f'{q} #|@ {a["text"]} #|@ {sent}\n')

    if args.templates == 'NA':
        stanza_lang = stanza.Pipeline(lang=args.lang, processors=dep_proc)
        templates, temp_fname = generate_templates(gen_fname, stanza_lang, min_support=args.min_support, strict=args.strict,
            rtl=args.right_to_left, case_folding=args.case_folding, remove_punct=args.remove_punct,
            temp_fname=os.path.join(DIR, "templates.txt"), sent_fname=os.path.join(DIR, "sentences.txt"),
            remove_diacritics=args.remove_diacritics, dot_fix=args.dot_fix, join_char=args.join_char,
            idf_file=args.idf)
    else:
        print("Loading templates from {}".format(args.templates))
        templates, temp_fname = templates = load_templates([args.templates]), args.templates

    guards = generate_guards(templates, temp_fname, guard_fname=os.path.join(DIR, "guards.txt")) if templates else []
    generation_time = round(time.time() - start, 2)
    print(f"Generation took {generation_time} seconds")

    if args.modeldb_url:
        run.log_observation("number_of_templates", len(templates))
        run.log_observation("number_of_guards", len(guards))
        run.log_observation("generation_time_in_seconds", generation_time)
        run.log_artifact("model", DIR)
    


import os
import csv
import sys
import unicodedata
import re
from collections import defaultdict
from pathlib import Path
from pprint import pprint
import logging

import numpy as np

COORDINATE_CLAUSES = ['conj']
SUBORDINATE_CLAUSES = ['csubj', 'xcomp', 'ccomp', 'advcl', 'acl', 'acl:relcl']

FUNCTION_WORDS = ['DET', 'PART', 'INTJ', 'SYM'] # AUX is a special case

MODIFIER_RELS = ["advmod", "amod"]

PUNCT_TABLE = dict.fromkeys(i for i in range(sys.maxunicode) if unicodedata.category(chr(i)).startswith('P'))

HOME_DIR = str(Path.home())
DEFAULT_TEMPLATES_DIR = os.getenv(
    'QUINDUCTOR_RESOURCES_DIR',
    os.path.join(HOME_DIR, 'quinductor_resources')
)
QUINDUCTOR_RESOURCES_GITHUB = 'https://raw.githubusercontent.com/dkalpakchi/quinductor/master/templates'

MODELS = {
    'ar': {
        'templates': 1614104416496133,
        'pos_ngrams': ['ar_padt_train.txt']
    },
    'en': {
        'templates': 16132054753040054,
        'pos_ngrams': ['ewt_train_freq', 'ewt_dev_freq']
    },
    'fi': {
        'templates': 16132078825085254,
        'pos_ngrams': ['fi_tdt_train.txt']
    },
    'id': {
        'templates': 16140609246000547,
        'pos_ngrams': ['id_gsd_train.txt']
    },
    'ja': {
        'templates': 16140572221308537,
        'pos_ngrams': ['ja_gsd_train.txt']
    },
    'ko': {
        'templates': 16140582210609627,
        'pos_ngrams': ['ko_gsd_train.txt']
    },
    'ru': {
        'templates': 1613204358381249,
        'pos_ngrams': ['ru_syntagrus_train.txt']
    },
    'te': {
        'templates': 16140691545631247,
        'pos_ngrams': ['te_mtg_train.txt']
    }
}


def get_logger():
    logger = logging.getLogger('quinductor')
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    console.setFormatter(formatter)
    logger.addHandler(console)
    return logger


def get_default_model_path(lang):
    if lang in MODELS:
        return os.path.join(DEFAULT_TEMPLATES_DIR, lang, str(MODELS[lang]['templates']))
    else:
        logger = logging.getLogger('quinductor')
        logger.error(
            """The language {} currently has no available models.
            Please create your own model and provide it using script arguments""".format(args.lang)
        )
        sys.exit(1)


class TemplateElement:
    def __init__(self, root_chain, node, last_span_id=None, is_subtree=False, is_lemma=False):
        self.__chain = root_chain
        self.__id = (node.id if last_span_id is None else last_span_id) if node else None
        self.__node = node
        self.__positives = []
        self.__negatives = []
        self.__is_subtree = is_subtree
        # the subtrees that can't be added definitely because the child of the node is already in the positives list
        self.__taboo = []
        self.__is_lemma = is_lemma

    @property
    def is_lemma(self):
        return self.__is_lemma

    def copy_positives(self, elements):
        for e in elements:
            self.__positives.extend(e.positives)
            self.__taboo.extend(e.taboo)

    def copy_negatives(self, elements):
        for e in elements:
            self.add_negatives(e.negatives)

    def add_positives(self, elements):
        for e in elements:
            self.__positives.append(e)
            if e[-1].endswith('*'):
                e_copy = list(e)
                e_copy[-1] = e_copy[-1][:-1]
                self.__taboo.append(e_copy)

            for k in range(len(e)):
                self.__taboo.append(e[:-1-k])

    def add_negatives(self, elements):
        # print("TABOO", self.__taboo)
        # print("POS", self.__positives)
        for e in elements:
            # print(e)
            # negatives list should be quite small (up to 10 elements),
            # so it should be quick to check `in` relationship
            if e not in self.__positives and e not in self.__negatives and e not in self.__taboo:
                self.__negatives.append(e)
        # print(self.__negatives)

    def merge_negatives(self, children_map):
        # TODO: wrong merging procedure
        # need to check if they can be merged in the first place
        if self.__negatives:
            # TODO: check if any of the values of the children map is fully in negatives
            # if so, swap all those elements for the corresponding key
            # NOTE: key is a str, so split it by '.' first
            set_negatives = set(map(lambda x: ".".join(x), self.__negatives))
            sorted_children = sorted(children_map.items(), reverse=True)
            for k, v in sorted_children:
                set_values = set(map(lambda x: ".".join(x), v))
                if set_values and (set_values & set_negatives) == set_values:
                    for node in v:
                        try:
                            self.__negatives.remove(node)
                        except ValueError:
                            # means we've already deleted this element as a part of other chain
                            pass
                    self.__negatives.append(k.split('.'))

    @property
    def positives(self):
        return self.__positives

    @property
    def negatives(self):
        return self.__negatives

    @property
    def taboo(self):
        return self.__taboo

    @property
    def node(self):
        return self.__node

    @property
    def chain(self):
        return self.__chain

    @property
    def id(self):
        return self.__id

    @property
    def is_subtree(self):
        return self.__is_subtree

    def __len__(self):
        return len(self.__chain)

    def __str__(self):
        prefix = '.'.join([x.replace('*', '') for x in self.__chain])
        inside = f"w{'.' + prefix if self.__chain else ''}"
        negate = " ".join([f"- {'.'.join(n).replace(prefix + '.', '')}" for n in self.__negatives])
        if negate:
            inside += " " + negate
        if self.__is_subtree:
            return f"<{inside}>"
        else:
            return f"[{inside}.lemma]" if self.__is_lemma else f"[{inside}]"

    def __and__(self, other):
        res = []
        for c in self.__chain:
            if c in other.chain:
                res.append(c)
            else:
                # the longest prefix subchain found
                break
        return res


def chain(node1, node2, include_ids=False, only_ids=False):
    """
    Find a chain of dependency tags from `node1` to `node2` (if possible)
    
    :param      node1:  The node 1
    :type       node1:  udon2.Node
    :param      node2:  The node 2
    :type       node2:  udon2.Node
    """
    node, chain = node2, []
    while not node.is_identical(node1, ""):
        chain.append(node.id if only_ids else f"{node.deprel}#{int(node.id)}" if include_ids else node.deprel)
        node = node.parent
    chain.reverse()
    return chain


def chain_to_root(node, include_ids=False, only_ids=False):
    n, chain = node, []
    while not n.is_root():
        chain.append(node.id if only_ids else f"{node.deprel}#{int(node.id)}" if include_ids else node.deprel)
        n = n.parent
    chain.reverse()
    return chain

def get_question_word(question_nodes, lang):
    N = len(question_nodes)
    start_word, next_index, cand = None, 1, question_nodes[0].form.lower()
    if cand in lang.OW_QUESTION_WORDS:
        start_word = cand

    j = 1
    while cand not in lang.MW_QUESTION_WORDS and j < N:
        cand += f" {question_nodes[j].form.lower()}"
        j += 1

    if j < N:
        start_word = cand
        next_index = j
    return start_word, next_index


def get_sentence_by_answer(answer, context, segmenter):
    # segmenter is a Stanza tokenizer
    doc = segmenter(context.strip().replace(u'\xa0', ' '))
    it = 0
    ans = answer['text'].strip().replace(u'\xa0', ' ')
    for s in doc.sentences:
        it += sys.getsizeof(s.text)
        sent = s.text

        if ans in sent and it >= answer['end']:
            for p in sent.strip().split('\n'):
                # TODO: some of the cases will be invalid, because of lists or "blankett nummer" cases
                if ans in p:
                    return p.strip()

def get_statistics(fname):
    qw_stat = defaultdict(int)
    with open(fname) as f:
        for line in f:
            if line.strip():
                question, answer, base_sentence = line.split("|")
                qw_stat[question.split()[0]] += 1
    return qw_stat


def get_difference(a, b):
    """
    Get the difference between `a` and `b` preserving the order of elements in `a`
    """
    if type(a) == str:
        a = a.split()
    if type(b) == str:
        b = b.split()

    diff = []
    for word in a:
        if word not in b:
            diff.append(word)
    return diff


def get_intersection(a, b):
    """
    Get the difference between `a` and `b` preserving the order of elements in `a`
    """
    if type(a) == str:
        a = a.split()
    if type(b) == str:
        b = b.split()

    if not a or not b:
        return []

    same = []
    for word in a:
        if word in b:
            same.append(word)
    return same


def check_multiword_qw(lst, lang):
    return " ".join(lst) in lang.MW_QUESTION_WORDS

def load_pos_ngrams(folder):
    bigrams_freq = defaultdict(lambda: defaultdict(int))
    trigrams_freq = defaultdict(lambda : defaultdict(lambda : defaultdict(int)))
    # morph_vocab = set()
    for fname in os.listdir(folder):
        definitions = True
        id2tag = {}
        with open(os.path.join(folder, fname)) as f:
            reader = csv.reader(f, delimiter=',')

            for line in reader:
                if line:
                    if definitions:
                        idx = line[0]
                        tag = ",".join(line[1:])
                        if '/' in tag:
                            pos, morph = tag.split('/')
                            morph_tags = sorted(morph.split('|'))
                            # for m in morph_tags:
                            #     morph_vocab.add(m)
                            id2tag[int(idx)] = f"{pos}/{'|'.join(morph_tags)}"
                        else:
                            id2tag[int(idx)] = tag
                    else:
                        if len(line) == 3:
                            w1, w2, freq = map(int, line)
                            bigrams_freq[id2tag[w1]][id2tag[w2]] += freq
                        elif len(line) == 4:
                            w1, w2, w3, freq = map(int, line)
                            trigrams_freq[id2tag[w1]][id2tag[w2]][id2tag[w3]] += freq
                else:
                    definitions = False
    
    unigrams_freq = defaultdict(int)    
    for w1 in bigrams_freq:
        for w2 in bigrams_freq[w1]:
            unigrams_freq[w1] += bigrams_freq[w1][w2]
            unigrams_freq[w2] += bigrams_freq[w1][w2]

    log_unigrams = defaultdict(float)
    log_S = np.log(sum(unigrams_freq.values()))
    for w1 in unigrams_freq:
        log_unigrams[w1] = np.log(unigrams_freq[w1]) - log_S

    log_bigrams = defaultdict(lambda: defaultdict(float))
    for w1 in bigrams_freq:
        log_S = np.log(sum(bigrams_freq[w1].values()))
        for w2 in bigrams_freq[w1]:
            log_bigrams[w1][w2] = np.log(bigrams_freq[w1][w2]) - log_S


    log_trigrams = defaultdict(lambda : defaultdict(lambda : defaultdict(float)))
    for w1 in trigrams_freq:
        for w2 in trigrams_freq[w1]:
            log_S = np.log(bigrams_freq[w1][w2])
            for w3 in trigrams_freq[w1][w2]:
                log_trigrams[w1][w2][w3] = np.log(trigrams_freq[w1][w2][w3]) - log_S

    return log_unigrams, log_bigrams, log_trigrams #, morph_vocab


def read_csv_eval(eval_fname):
    sentences, hypothesis, answers, written = [], [], [], False
    scores, score_sum, total = [], 0, 0
    with open(eval_fname) as f:
        reader = csv.reader(f, delimiter='|')
        headers = next(reader)
        for row in reader:
            if row:
                score_sum += float(row[4])
                total += 1
                if not written:
                    sentences.append(row[1])
                    hypothesis.append(row[2])
                    answers.append(row[3])
                    scores.append(float(row[4]))
                    written = True
            else:
                written = False
    return {
        'hypothesis': hypothesis,
        'sentences': sentences,
        'answers': answers,
        'scores': scores,
        'threshold': score_sum / total
    }


def repr_pos_morph(w):
    if type(w) == str:
        return w

    if w.feats:
        # str(w.feats) should be sorted
        token = f"{w.upos}/{str(w.feats)}"
    else:
        token = w.upos
    return token

def remove_unicode_punctuation(text):
    return text.translate(PUNCT_TABLE)

def is_punctuation(x):
    return unicodedata.category(x).startswith('P')


def remove_unicode_diacritics(text):
    return ''.join(c for c in unicodedata.normalize('NFD', text)
        if unicodedata.category(c) != 'Mn')

def load_idf(fname):
    idf = {}
    with open(fname, 'r') as f:
        reader = csv.reader(f)
        for line in reader:
            try:
                word, v = line
                idf[word] = float(v)
            except:
                continue
    return idf


import setuptools

setuptools.setup(
    name='quinductor',
    version='0.1.0',    
    description='A package implementing a multi-lingual question generation method described in https://arxiv.org/abs/2103.10121',
    url='https://github.com/dkalpakchi/quinductor',
    author='Dmytro Kalpakchi',
    install_requires=[
        'numpy',
        'udon2',
        'pypeg2',
        'stanza',
        'tqdm',
        'dill',
        'jsonlines'
    ],
    classifiers=[
        'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    package_dir={"": "src"},
    package_data={
        'quinductor': ['lang_spec_feats.json'],
    },
    packages=setuptools.find_packages(where="src")
)
